// Written by Robert C. Kirby
// Copyright 2005 by The University of Chicago
// Distributed under the LGPL license
// This work is partially supported by the US Department of Energy
// under award number DE-FG02-04ER25650

#include "FIAT_expansions.hpp"
#include "FIAT_jacobi.hpp"

using namespace FIAT;
using namespace blitz;

double Expansions::psitilde_a( int i , double z )
{
  return Jacobi::eval_jacobi(0,0,i,z);
}

double Expansions::psitilde_b( int i, int j , double z )
{
  if (i==0) {
    return Jacobi::eval_jacobi(1,0,j,z);
  }
  else {
    return pow( 0.5 * (1.0-z) , i ) * Jacobi::eval_jacobi(2*i+1,0,j,z);
  }
}

double Expansions::psitilde_c(int i,int j,int k,double z)
{
  if (i+j==0){
    return Jacobi::eval_jacobi(2,0,k,z);
  }
  else{
    return pow( 0.5 * (1.0-z) , i+j )
      * Jacobi::eval_jacobi(2*(i+j+1),0,k,z);
  }
}

Array<double,2>
Expansions::make_scalings(int n, const Array<double,1> etas )
{
  Array<double,2> scalings(n+1,etas.rows());

  Array<double,1> rzero = scalings(0,Range::all());
  scalings = 1.0;

  if (n>0) {
    Array<double,1> rone = scalings(1,Range::all());
    rone = 0.5 * (1.0 - etas);
    for (int k=2;k<n+1;k++) {
      Array<double,1> rk = scalings(k,Range::all());
      Array<double,1> rkm1 = scalings(k-1,Range::all());
      rk = rkm1 * rone;
    }
  }

  return scalings;
}

static Array<double,1>
eta_triangle( const Array<double,1> &xi )
{
  Array<double,1> eta(2);
  eta(0) = (xi(1) == 1.0)?-1.0:2.0*(1.0+xi(0))/(1.0-xi(1))-1.0;
  eta(1) = xi(1);
  return eta;
}

static Array<double,1>
xi_triangle( const Array<double,1> &eta )
{
  Array<double,1> xi(2);
  xi(0) = 0.5 * (1.0 + eta(0))*(1.0-eta(1))-1.0;
  xi(1) = eta(1);
  return xi;
}

static Array<double,1>
eta_tetrahedron(const Array<double,1> &xi )
{

  Array<double,1> eta(3);
  eta(0) = (xi(1)+xi(2)==0.0)?1.0:(-2.*(1.+xi(0))/(xi(1)+xi(2))-1.);
  eta(1) = (xi(2)==1.0)?-1.0:(2.0*(1.+xi(1))/(1.-xi(2))-1.);
  eta(2) = xi(2);

  return eta;
}

static Array<double,1>
xi_tetrahedron(const Array<double,1> &eta )
{
  Array<double,1> xi(3);
  xi(0)=0.25*(1.0+eta(0))*(1.0-eta(1))*(1.0-eta(2))-1.0;
  xi(1)=0.5*(1.0+eta(1))*(1.0-eta(2))-1.0;
  xi(2)=eta(2);
  return xi;
}

static Array<double,1> do_nothing( const Array<double,1> &x)
{
  return x; 
}

static coord_transform cc[] = {NULL ,
			       do_nothing ,
			       eta_triangle ,
			       eta_tetrahedron};
static coord_transform icc[] = {NULL , 
				do_nothing,
				xi_triangle,
				xi_tetrahedron};


coord_transform Expansions::coord_change( const int shape )
{
  if (shape < 1 || shape > 3) FIAT_ERROR("invalid shape code " << shape 
                                         << " in Expansions::coord_change");
  return cc[shape];
}

coord_transform 
Expansions::inverse_coord_change( const int shape )
{
  if (shape < 1 || shape > 3) FIAT_ERROR("invalid shape code " << shape 
                                         << " in Expansions::inverse_coord_change");
  return icc[shape];
}


// Tabulates all the basis functions over the lines up
// to degree n at all points xs
// returns an array result[i][j] = phi_i(x_j)
static Array<double,2>
tabulate_phis_line( int n ,
		    const Array<double,2> &xs )
{
  Array<double,1> ys = xs(Range::all() , 0 );


  Array<double,2> phis = Jacobi::eval_jacobi_batch(0,0,n,ys);

  for (int i=0;i<phis.extent(0);i++) {
    double alpha = sqrt( 1.0 * i + 0.5 );
    Array<double,1> row_i = phis(i,Range::all());
    row_i *= alpha;
  }

  return phis;
}

// Tabulates all the basis functions over the lines up
// to degree n at all points xs
// returns an array dphis(i,j,k) = d_dxi phi_j (x_k)

static Array<double,3>
tabulate_phis_derivs_line( int n ,
			   const Array<double,2> &xs )
{
  Array<double,1> ys = xs( Range::all() , 0 );

  Array<double,2> dphis_inside = Jacobi::eval_jacobi_deriv_batch(0,0,n,ys);

  Array<double,3> dphis( dphis_inside.data() , shape(1,n+1,xs.rows()) ,
			 duplicateData );

  for (int i=0;i<dphis_inside.extent(0);i++) {
    Array<double,1> row_i = dphis(0,i,Range::all());
    row_i *= sqrt(1.0*i+0.5);
  }

  return dphis;
}

static Array<double,2>
tabulate_phis_triangle( int n ,
			const Array<double,2> &xs )
{
  int num_poly = Shapes::polynomial_dimension( Shapes::TRIANGLE , n );
  Array<double,2> phis(num_poly,xs.rows());
  Array<double,2> etas(xs.shape());

  for (int i=0;i<xs.extent(0);i++) {
    coord_transform f = Expansions::coord_change( Shapes::TRIANGLE );
    etas(i,Range::all()) = f(xs(i,Range::all()));
  }

  Array<double,1> eta1s = etas(Range::all(),0);
  Array<double,1> eta2s = etas(Range::all(),1);

  Array<double,2> psitilde_as = Jacobi::eval_jacobi_batch(0,0,n,eta1s);
  Array<double,2> scalings = Expansions::make_scalings(n,eta2s);

  vector<Array<double,2> > psitilde_bs;

  for (int i=0;i<n+1;i++) {
    Array<double,2> psitilde_bs_cur =
      Jacobi::eval_jacobi_batch(2*i+1,0,n-i,eta2s);
    psitilde_bs.push_back(psitilde_bs_cur);
  }

  int cur = 0;
  for (int k=0;k<n+1;k++) {
    for (int i=0;i<k+1;i++) {
      int ii=k-i;
      int jj=i;
      double alpha = sqrt((ii+0.5)*(ii+jj+1.0));
      Array<double,1> phis_cur = phis(cur,Range::all());
      Array<double,1> psitilde_as_ii = psitilde_as(ii,Range::all());
      Array<double,1> scalings_ii = scalings(ii,Range::all());
      Array<double,2> psitilde_bs_ii = psitilde_bs[ii];
      Array<double,1> psitilde_bs_ii_jj = psitilde_bs_ii(jj,Range::all());
      
      phis_cur = alpha * psitilde_as_ii * scalings_ii * psitilde_bs_ii_jj;

      cur++;
    }
  }

  return phis;
}

// can't do derivative at singular point on the top.
static Array<double,3>
tabulate_phis_derivs_triangle( int n,
			       const Array<double,2> & xs )
{
  int num_poly = Shapes::polynomial_dimension( Shapes::TRIANGLE , n );
  int num_points = xs.extent(0);
  Array<double,3> dphis(2,num_poly,num_points);

  // transform points
  Array<double,2> etas( xs.shape() );
  
  for (int i=0;i<xs.extent(0);i++) {
    coord_transform f = Expansions::coord_change( Shapes::TRIANGLE );
    etas(i,Range::all()) = f( xs(i,Range::all()) );
  } 

  Array<double,1> eta1s = etas( Range::all() , 0 );
  Array<double,1> eta2s = etas( Range::all() , 1 );

  // tabulate principal basis functions
  Array<double,2> psitilde_as = Jacobi::eval_jacobi_batch(0,0,n,eta1s);
  Array<double,2> psitilde_as_derivs = 
    Jacobi::eval_jacobi_deriv_batch(0,0,n,eta1s);

  vector<Array<double,2> > psitilde_bs;
  vector<Array<double,2> > psitilde_bs_derivs;

  for (int i=0;i<n+1;i++) {
    Array<double,2> psitilde_bs_cur = 
      Jacobi::eval_jacobi_batch(2*i+1,0,n-i,eta2s);
    Array<double,2> psitilde_bs_derivs_cur =
      Jacobi::eval_jacobi_deriv_batch(2*i+1,0,n-i,eta2s);

    psitilde_bs.push_back( psitilde_bs_cur );
    psitilde_bs_derivs.push_back( psitilde_bs_derivs_cur );
  }

  // tabulate powers of (0.5*(1-eta2))**i
  Array<double,2> scalings = Expansions::make_scalings(n,eta2s);

  Array<double,2> xderivs = dphis(0,Range::all(),Range::all());
  Array<double,2> yderivs = dphis(1,Range::all(),Range::all());

  int cur = 0;
  for (int k=0;k<n+1;k++) {
    for (int i=0;i<k+1;i++) {
      int ii=k-i;
      int jj=i;

      double alpha = sqrt( (ii+0.5)*(ii+jj+1.0) );

      Array<double,1> xderivs_cur = xderivs(cur,Range::all());
      Array<double,1> yderivs_cur = yderivs(cur,Range::all());

      Array<double,1> psitilde_as_cur = psitilde_as(ii,Range::all());
      Array<double,1> psitilde_as_derivs_cur 
	= psitilde_as_derivs(ii,Range::all());

      Array<double,1> psitilde_bs_cur = psitilde_bs[ii](jj,Range::all());
      Array<double,1> psitilde_bs_derivs_cur
	= psitilde_bs_derivs[ii](jj,Range::all());

      Array<double,1> scalings_cur = scalings(ii,Range::all());
      Array<double,1> scalings_cur_m1 = scalings(ii-1,Range::all());

      Array<double,1> tmp(num_points);

      xderivs_cur = psitilde_as_derivs_cur * psitilde_bs_cur;
      if (ii>0) xderivs_cur *= scalings_cur_m1;

      yderivs_cur = psitilde_as_derivs_cur
	* psitilde_bs_cur
	* 0.5 * (1.0 + eta1s);

      if (ii>0) yderivs_cur *= scalings_cur_m1;

      tmp = psitilde_bs_derivs_cur * scalings_cur;
      if (ii>0) tmp -= 0.5 * ii * psitilde_bs_cur * scalings_cur_m1;

      yderivs_cur += psitilde_as_cur * tmp;

      xderivs_cur *= alpha;
      yderivs_cur *= alpha;
     
      cur++;
    }
  }

  return dphis;

}


Array<double,2>
tabulate_phis_tetrahedron( int n ,
			   const Array<double,2> &xs )
{
  int num_poly = Shapes::polynomial_dimension( Shapes::TETRAHEDRON , n );
  int num_points = xs.extent(0);
  Array<double,2> phis(num_poly,num_points);

  // transform points
  
  Array<double,2> etas( xs.shape() );

  coord_transform f = Expansions::coord_change( Shapes::TETRAHEDRON );
  for (int i=0;i<xs.extent(0);i++) {
    etas(i,Range::all()) = f( xs(i,Range::all()) );
  } 

  Array<double,1> eta1s = etas( Range::all() , 0 );
  Array<double,1> eta2s = etas( Range::all() , 1 );
  Array<double,1> eta3s = etas( Range::all() , 2 );

  Array<double,2> psitilde_as = 
    Jacobi::eval_jacobi_batch(0,0,n,eta1s);

  Array<double,2> psitilde_as_derivs = 
    Jacobi::eval_jacobi_deriv_batch(0,0,n,eta1s);

  vector<Array<double,2> > psitilde_bs;

  for (int i=0;i<n+1;i++) {
    Array<double,2> psitilde_bs_cur = 
      Jacobi::eval_jacobi_batch(2*i+1,0,n-i,eta2s);
    psitilde_bs.push_back( psitilde_bs_cur );
  }

  vector<vector<Array<double,2> > > psitilde_cs;

  for (int i=0;i<n+1;i++) {
    vector<Array<double,2> > psitilde_cs_i;
    vector<Array<double,2> > psitilde_cs_derivs_i;
    for (int j=0;j<n-i+1;j++) {
      Array<double,2> psitilde_cs_i_j =       
	Jacobi::eval_jacobi_batch(2*(i+j+1),0,n-i-j,eta3s);
      psitilde_cs_i.push_back( psitilde_cs_i_j );
    }
    psitilde_cs.push_back( psitilde_cs_i );
  }


  Array<double,2> eta2_scalings = Expansions::make_scalings(n,eta2s);
  Array<double,2> eta3_scalings = Expansions::make_scalings(n,eta3s);

  int cur = 0;
  for (int k=0;k<n+1;k++) {
    for (int i=0;i<k+1;i++) {
      for (int j=0;j<k-i+1;j++) {
	int ii = k-i-j;
	int jj = j;
	int kk = i;

	double alpha = sqrt( (ii+0.5) * (ii+jj+1.0) * (ii+jj+kk+1.5) );

	Array<double,1> phis_cur = phis(cur,Range::all());
	Array<double,1> psitilde_as_ii = psitilde_as(ii,Range::all());
	Array<double,1> psitilde_bs_ii_jj = psitilde_bs[ii](jj,Range::all());
	Array<double,1> psitilde_cs_ii_jj_kk = psitilde_cs[ii][jj](kk,Range::all());
	Array<double,1> eta2_scalings_ii = eta2_scalings(ii,Range::all());
	Array<double,1> eta3_scalings_iipjj = eta3_scalings(ii+jj,Range::all());

	phis_cur = psitilde_as_ii
	  * eta2_scalings_ii
	  * psitilde_bs_ii_jj
	  * eta3_scalings_iipjj
	  * psitilde_cs_ii_jj_kk
	  * alpha;


	cur++;
      }
    }
  }

  return phis;
}

static Array<double,3>
tabulate_phis_derivs_tetrahedron( int n ,
				  const Array<double,2> &xs )
{
  int num_poly = Shapes::polynomial_dimension( Shapes::TETRAHEDRON , n );
  int num_points = xs.extent(0);
  Array<double,3> dphis(3,num_poly,num_points);

  // transform points
  
  Array<double,2> etas( xs.extent(0) , xs.extent(1) );

  coord_transform f = Expansions::coord_change( Shapes::TETRAHEDRON );
  for (int i=0;i<xs.extent(0);i++) {
    etas(i,Range::all()) = f( xs( i,Range::all() ) );
  } 

  Array<double,1> eta1s = etas( Range::all() , 0 );
  Array<double,1> eta2s = etas( Range::all() , 1 );
  Array<double,1> eta3s = etas( Range::all() , 2 );

  Array<double,2> psitilde_as = 
    Jacobi::eval_jacobi_batch(0,0,n,eta1s);

  Array<double,2> psitilde_as_derivs = 
    Jacobi::eval_jacobi_deriv_batch(0,0,n,eta1s);

  vector<Array<double,2> > psitilde_bs;
  vector<Array<double,2> > psitilde_bs_derivs;

  for (int i=0;i<n+1;i++) {
    Array<double,2> psitilde_bs_cur =
      Jacobi::eval_jacobi_batch(2*i+1,0,n-i,eta2s);
    Array<double,2> psitilde_bs_derivs_cur =
      Jacobi::eval_jacobi_deriv_batch(2*i+1,0,n-i,eta2s);

    psitilde_bs.push_back( psitilde_bs_cur );
    psitilde_bs_derivs.push_back( psitilde_bs_derivs_cur );
  }

  vector<vector<Array<double,2> > > psitilde_cs;
  vector<vector<Array<double,2> > > psitilde_cs_derivs;

  for (int i=0;i<n+1;i++) {
    vector<Array<double,2> > psitilde_cs_i;
    vector<Array<double,2> > psitilde_cs_derivs_i;
    for (int j=0;j<n-i+1;j++) {
      Array<double,2> psitilde_cs_i_j=
	Jacobi::eval_jacobi_batch(2*(i+j+1),0,n-i-j,eta3s);
      Array<double,2> psitilde_cs_derivs_i_j =
	Jacobi::eval_jacobi_deriv_batch(2*(i+j+1),0,n-i-j,eta3s);

      psitilde_cs_i.push_back( psitilde_cs_i_j );
      psitilde_cs_derivs_i.push_back( psitilde_cs_derivs_i_j );
    }

    psitilde_cs.push_back( psitilde_cs_i );
    psitilde_cs_derivs.push_back( psitilde_cs_derivs_i );
  }

  Array<double,2> eta2_scalings =
    Expansions::make_scalings(n,eta2s);
  Array<double,2> eta3_scalings =
    Expansions::make_scalings(n,eta3s);

  Array<double,2> xderivs = dphis(0,Range::all(),Range::all());
  Array<double,2> yderivs = dphis(1,Range::all(),Range::all());
  Array<double,2> zderivs = dphis(2,Range::all(),Range::all());
  
  Array<double,1> tmp( num_points );

  int cur = 0;
  for (int k=0;k<n+1;k++) {
    for (int i=0;i<k+1;i++) {
      for (int j=0;j<k-i+1;j++) {
	int ii = k-i-j;
	int jj = j;
	int kk = i;

	double alpha = sqrt( (ii+0.5) * (ii+jj+1.0) * (ii+jj+kk+1.5) );

	Array<double,1> xderivs_cur = xderivs(cur,Range::all());
	Array<double,1> yderivs_cur = yderivs(cur,Range::all());
	Array<double,1> zderivs_cur = zderivs(cur,Range::all());

	Array<double,1> psitilde_as_ii = psitilde_as(ii,Range::all());
	Array<double,1> psitilde_as_derivs_ii
	  = psitilde_as_derivs(ii,Range::all());

	Array<double,1> psitilde_bs_ii_jj = psitilde_bs[ii](jj,Range::all());
	Array<double,1> psitilde_bs_derivs_ii_jj 
	  = psitilde_bs_derivs[ii](jj,Range::all());

	Array<double,1> psitilde_cs_ii_jj_kk 
	  = psitilde_cs[ii][jj](kk,Range::all());

	Array<double,1> psitilde_cs_derivs_ii_jj_kk 
	  = psitilde_cs_derivs[ii][jj](kk,Range::all());

	Array<double,1> eta2_scalings_ii = eta2_scalings(ii,Range::all());
	Array<double,1> eta2_scalings_iim1 = eta2_scalings(ii-1,Range::all());
	Array<double,1> eta3_scalings_iipjj 
	  = eta3_scalings(ii+jj,Range::all());
	Array<double,1> eta3_scalings_iipjjm1
	  = eta3_scalings(ii+jj-1,Range::all());
	
	xderivs_cur = psitilde_as_derivs_ii
	  * psitilde_bs_ii_jj
	  * psitilde_cs_ii_jj_kk;
	if (ii>0) xderivs_cur *= eta2_scalings_iim1;
	if (ii+jj>0) xderivs_cur *= eta3_scalings_iipjjm1;

// 	yderivs_cur = psitilde_as_derivs_ii
// 	  * psitilde_bs_ii_jj
// 	  * psitilde_cs_ii_jj_kk
// 	  * 0.5 * (1.0+eta1s);
// 	if (ii>0) yderivs_cur *= eta2_scalings_iim1;
// 	if (ii+jj>0) zderivs_cur *= eta3_scalings_iipjjm1;

// 	tmp = psitilde_bs_derivs_ii_jj
// 	  * eta2_scalings_ii;
// 	if (ii>0) tmp -= 0.5*ii*eta2_scalings_iim1 * psitilde_bs_ii_jj;
// 	tmp *= psitilde_as_ii * psitilde_cs_ii_jj_kk;
// 	if (ii+jj>0) tmp *= eta3_scalings_iipjjm1;
	
// 	yderivs_cur += tmp;
	//d/deta1 with scalings
	yderivs_cur = psitilde_as_derivs_ii;
	yderivs_cur *= psitilde_bs_ii_jj;
	yderivs_cur *= psitilde_cs_ii_jj_kk;
	yderivs_cur *= 0.5 * (1.0+eta1s);
	if (ii>0) {
	  yderivs_cur *= eta2_scalings_iim1;
	}
	if (ii+jj>0){
	  yderivs_cur *= eta3_scalings_iipjjm1;
	}
	//tmp will hold d/deta2 term
	tmp = psitilde_bs_derivs_ii_jj;
	tmp *= eta2_scalings_ii;
	if (ii>0){
	  tmp -= 0.5*ii*eta2_scalings_iim1*psitilde_bs_ii_jj;
	}
	tmp *= psitilde_as_ii;
	tmp *= psitilde_cs_ii_jj_kk;
	if (ii+jj>0){
	  tmp *= eta3_scalings_iipjjm1;
	}
	yderivs_cur += tmp;

	zderivs_cur = psitilde_as_derivs_ii
	  * psitilde_bs_ii_jj
	  * psitilde_cs_ii_jj_kk
	  * 0.5 * (1.0+eta1s);
	if (ii>0) zderivs_cur *= eta2_scalings_iim1;
	if (ii+jj>0) zderivs_cur *= eta3_scalings_iipjjm1;

	tmp = psitilde_bs_derivs_ii_jj
	  * eta2_scalings_ii;
	if (ii>0) tmp -= 0.5*ii*eta2_scalings_iim1*psitilde_bs_ii_jj;
	tmp *= psitilde_as_ii * psitilde_cs_ii_jj_kk * 0.5 * (1.0+eta2s);
	if (ii+jj>0) tmp *= eta3_scalings_iipjjm1;
	zderivs_cur += tmp;

	tmp = psitilde_cs_derivs_ii_jj_kk
	  * eta3_scalings_iipjj;
	if (ii+jj>0) tmp -= 0.5*(ii+jj)*psitilde_cs_ii_jj_kk*eta3_scalings_iipjjm1;
	tmp *= psitilde_as_ii * psitilde_bs_ii_jj * eta2_scalings_ii;
	
	zderivs_cur += tmp;
	

	xderivs_cur *= alpha;
	yderivs_cur *= alpha;
	zderivs_cur *= alpha;

	cur++;
      }
    }
  }

  return dphis;
}

static tabulator tabulators[] 
= {NULL,
   tabulate_phis_line,
   tabulate_phis_triangle,
   tabulate_phis_tetrahedron};

static deriv_tabulator deriv_tabulators[] 
= {NULL,
   tabulate_phis_derivs_line,
   tabulate_phis_derivs_triangle,
   tabulate_phis_derivs_tetrahedron};


tabulator Expansions::get_tabulator( int shp )
{
  if (shp < 1 || shp > 3) FIAT_ERROR("invalid shape code " << shp 
                                     << " in Expansions::get_tabulator");
  return tabulators[shp];
}

deriv_tabulator Expansions::get_deriv_tabulator( int shp )
{
  if (shp < 1 || shp > 3)  FIAT_ERROR("invalid shape code " << shp 
                                      << " in Expansions::get_deriv_tabulator");
  return deriv_tabulators[shp];
}

