// Written by Robert C. Kirby
// Copyright 2005 by The University of Chicago
// Distributed under the LGPL license
// This work is partially supported by the US Department of Energy
// under award number DE-FG02-04ER25650

#include "FIAT_jacobi.hpp"

using namespace FIAT;
using namespace blitz;

// evaluates Jacobi polynomial at a point x
// recurrence relations are implemented from the pseudocode
// given in Karniadakis & Sherwin, Appendix a
double Jacobi::eval_jacobi( double a, double b, int n, double x )
{
  if ( 0 == n ) {
    return 1.0;
  }
  else if ( 1 == n ) {
    return 0.5 * ( a - b + ( a + b + 2.0 ) * x );
  }
  else { // 2 <= n
    double apb = a + b;
    double pn2 = 1.0;
    double pn1 = 0.5 * ( a - b + ( apb + 2.0 ) * x );
    double p = 0.0;
    for (int k=2;k<=n;k++) {
      double a1 = 2.0 * k * ( k + apb ) * ( 2.0 * k + apb - 2.0 );
      double a2 = ( 2.0 * k + apb - 1.0 ) * ( a * a - b * b );
      double a3 = ( 2.0 * k + apb - 2.0 )
	* ( 2.0 * k + apb - 1.0 )
	* ( 2.0 * k + apb );
      double a4 = 2.0 * ( k + a - 1.0 ) * ( k + b - 1.0 ) 
	* ( 2.0 * k + apb );
      a2 /= a1;
      a3 /= a1;
      a4 /= a1;
      p = ( a2 + a3 * x ) * pn1 - a4 * pn2;
      pn2 = pn1;
      pn1 = p;
    }
    return p;
  }
}

// Evaluates all jacobi polynomials with weights a,b up to
// degree n.  xs is a vector of points.  Returns a 2-dimensional
// array, where the rows index is the number of the Jacobi polynomial
// and the column index is the number of the point
Array<double,2> 
Jacobi::eval_jacobi_batch( double a , double b , 
			   int n , 
			   const Array<double,1> &xs )
{
  Array<double,2> result(n+1,xs.extent(0));

  result(0,Range::all()) = 1.0;

  if (n>0) {
    double apb = a+b;
    result(1,Range::all()) = 0.5 * (a-b+(apb+2.) * xs);

    for (int k=2;k<n+1;k++) {
      double a1 = 2.0 * k * ( k + apb ) * ( 2.0 * k + apb - 2.0 );
      double a2 = ( 2.0 * k + apb - 1.0 ) * ( a * a - b * b );
      double a3 = ( 2.0 * k + apb - 2.0 )
	* ( 2.0 * k + apb - 1.0 )
	* ( 2.0 * k + apb );
      double a4 = 2.0 * ( k + a - 1.0 ) * ( k + b - 1.0 ) 
	* ( 2.0 * k + apb );
      a2 /= a1;
      a3 /= a1;
      a4 /= a1;

      result(k,Range::all()) = (a2+a3*xs)*result(k-1,Range::all())
	- a4 * result(k-2,Range::all());
    }
  }

  return result;
}


double 
Jacobi::eval_jacobi_deriv( double a , double b , int n , double x )
{
  if ( 0 == n ) {
    return 0.0;
  }
  else {
    return 0.5 * ( a + b + n + 1 ) * eval_jacobi( a+1 , b+1 , n-1 , x );
  }
}

Array<double,2>
Jacobi::eval_jacobi_deriv_batch( double a , double b , int n ,
				 const Array<double,1> &xs )
{
  Array<double,2> result(n+1,xs.extent(0));

  result(0,Range::all()) = 0.0;

  if (0 == n) {
    return result;
  }
  else {
    result(Range(1,n),Range::all()) = eval_jacobi_batch(a+1,b+1,n-1,xs);
    for (int k=1;k<n+1;k++) {
      result(k,Range::all()) *= 0.5 * (a+b+k+1);
    }		      
    return result;
  }
}

