// Written by Robert C. Kirby
// Copyright 2005 by The University of Chicago
// Distributed under the LGPL license
// This work is partially supported by the US Department of Energy
// under award number DE-FG02-04ER25650

#include "FIAT_polynomial.hpp"
#include <iostream>

using namespace FIAT;
using namespace blitz;

PolynomialBase::PolynomialBase( const PolynomialBase & PB ) :
  shp( PB.shp ) , n( PB.n ) , dmats( PB.dmats ) {}

PolynomialBase::PolynomialBase( int shp , int n ) :
  shp( shp ) , n( n )
{
  int td = Shapes::topological_dimension( shp );
  int pd = Shapes::polynomial_dimension( shp , n );

  dmats.resize(td,pd,pd);

  // vandermonde matrix
  Array<double,2> xs = Shapes::make_points(shp,td,0,n+td+1);
  Array<double,2> vinv = Expansions::get_tabulator( shp )( n , xs );

  invert( vinv );

  Array<double,3> dtildes = 
    Expansions::get_deriv_tabulator(shp)(n,xs);
  

  for (int ii=0;ii<td;ii++) {
    Array<double,2> dtilde_cur = dtildes(ii,Range::all(),Range::all());
    Array<double,2> dmat_cur = dmats(ii,Range::all(),Range::all());
    mat_mat_mul( dtilde_cur , vinv , dmat_cur );
  }


}

Array<double,2> PolynomialBase::tabulate( const Array<double,2> &xs ) const
{
  return Expansions::get_tabulator( shp )( n , xs );
}

Array<double,1> PolynomialBase::eval_all( const Array<double,1> &x ) const
{
  Array<double,2> xs(1,x.rows());
  xs( 0 , Range::all() ) = x;
  Array<double,2> phis = tabulate( xs );
  return phis(Range::all(),0);
}

ScalarPolynomial::ScalarPolynomial( const ScalarPolynomial &sp ) :
  PB( sp.PB ) , dof( sp.dof ) {}

ScalarPolynomial::ScalarPolynomial( RCP<PolynomialBase> PB ,
				    Array<double,1> dof ) :
  PB( PB ) , dof( dof ) {}

double ScalarPolynomial::operator()( const Array<double,1> &x )
{
  Array<double,1> phis = PB->eval_all( x );
  return sum( dof * phis );
}

RCP<ScalarPolynomial> ScalarPolynomial::deriv( int dir ) const
{

  if (0 > dir || dir >= PB->topological_dimension() ) 
    FIAT_ERROR("invalid direction " << dir << " in ScalarPolynomial::deriv()");
  
  Array<double,1> deriv_dof(dof.size());
  Array<double,2> dmat = PB->dmat( dir );
  firstIndex i;
  secondIndex j;

  deriv_dof = sum( dmat( i , j ) * dof( j ) , j );

  RCP<ScalarPolynomial> poly = new ScalarPolynomial(PB , deriv_dof );
  return poly;
}

ScalarPolynomialSet::ScalarPolynomialSet( RCP<PolynomialBase> PB ,
					  Array<double,2> coeffs ) :
  PB(PB) , coeffs(coeffs)
{
}


Array<double,2> 
ScalarPolynomialSet::tabulate( const Array<double,2> &xs ) const
{
  Array<double,2> phis = PB->tabulate(xs);
  Array<double,2> vals(coeffs.rows(),phis.columns());

  // mat_mat_mul( coeffs , phis , vals );
  
  firstIndex i;
  secondIndex j;
  thirdIndex k;

  vals = sum( coeffs(i,k) * phis(k,j) , k );

  return vals;
}

Array<double,2>
ScalarPolynomialSet::trace_tabulate( int d , int e , 
                                     Array<double,2> pre_xs ) const
{
  PtMap &f = Shapes::get_pt_map(domain_shape(),d,e);
  Array<double,2> xs(pre_xs.rows(),topological_dimension());
  for (int i=0;i<pre_xs.rows();i++) {
    Array<double,1> xcur = xs(i,Range::all());
    Array<double,1> pre_xcur = pre_xs(i,Range::all());
    f( pre_xcur , xcur );
  }

  return tabulate( xs );
}

Array<double,1> ScalarPolynomialSet::eval_all( const Array<double,1> &x ) const
{
  Array<double,2> xs(1,x.size());
  xs(0,Range::all()) = x;
  Array<double,2> vals = tabulate(xs);
  return vals(Range::all(),0);
}

RCP<ScalarPolynomialSet> ScalarPolynomialSet::deriv_all( int dir ) const
{
  if (dir < 0 || dir >= topological_dimension()) 
    FIAT_ERROR("invalid direction " << dir 
               << " in ScalarPolynomial::deriv_all()");

  Array<double,2> dmat = PB->dmat( dir );
  Array<double,2> new_coeffs( coeffs.extent(0) , coeffs.extent(1) );

  firstIndex i;
  secondIndex j;
  thirdIndex k;

  new_coeffs = sum( coeffs(i,k) * dmat(j,k) , k );

  return RCP<ScalarPolynomialSet>(new ScalarPolynomialSet(PB,new_coeffs));
}

RCP<ScalarPolynomialSet>
ScalarPolynomialSet::multi_deriv_all( const Array<int,1> &alpha ) const
{
  if (alpha.size() != topological_dimension()) 
    FIAT_ERROR("invalid multiindex " << alpha 
               << " in ScalarPolynomial::deriv_all()");
  Array<double,2> new_coeffs(coeffs.extent(0),coeffs.extent(1));
  Array<double,2> tmp(coeffs.extent(0),coeffs.extent(1));

  new_coeffs = coeffs;

  for (int ii=0;ii<alpha.size();ii++){
    Array<double,2> dmat = PB->dmat( ii );
    for (int jj=0;jj<alpha(ii);jj++) {
      firstIndex i;
      secondIndex j;
      thirdIndex k;
      
      tmp = sum( new_coeffs(i,k) * dmat(j,k) , k );
      new_coeffs = tmp;
    }
  }

  return RCP<ScalarPolynomialSet>( new ScalarPolynomialSet( PB , 
							    new_coeffs ) );
  
}


RCP<ScalarPolynomial> ScalarPolynomialSet::operator[](int i)
{
  if (i<0 || i>= size()) 
    FIAT_ERROR("invalid index " << i
               << " in ScalarPolynomial::operator[]()");
  Array<double,1> dof = coeffs(i,Range::all());
  RCP<ScalarPolynomial> poly = new ScalarPolynomial(PB,dof);
  return poly;
}

namespace FIAT {
Array<double,2> identity( int n )
{
  Array<double,2> I(n,n);
  I = 0.0;
  for (int i=0;i<n;i++) I(i,i) = 1.0;
  return I;
}
}

