// Written by Robert C. Kirby
// Copyright 2005 by The University of Chicago
// Distributed under the LGPL license
// This work is partially supported by the US Department of Energy
// under award number DE-FG02-04ER25650


#include "FIAT_quadrature.hpp"

using namespace FIAT;
using namespace blitz;


static int factorial( int n )
{
  if (n<0) FIAT_ERROR("invalid argument " << n << " to factorial()");
  int result = 1;
  for (int i=1;i<n+1;i++) {
    result *= i;
  }
  return result;
}

// computes the m roots of P_{m}^{a,b} on [-1,1] by Newton's method
// using initial guesses of Chebyshev points, as in Karniadakis & Sherwin
static Array<double,1>
gauss_jacobi_points( double a, double b, int m )
{
  Array<double,1> xs(m);

  double eps = 1.0e-8;
  int max_iter = 100;
  for (int k=0;k<m;k++) {
    double r = -cos( (2.0*k+1.0)*M_PI / (2.0*m) );
    if (k>0) {
      r = 0.5 *( r + xs(k-1) );
    }
    int j=0;
    double delta = 2.0 * eps;
    while (j<max_iter) {
      double s=0.0;
      for (int i=0;i<k;i++) {
	s += 1.0 / (r-xs(i));
      }
      double f = Jacobi::eval_jacobi(a,b,m,r);
      double fp = Jacobi::eval_jacobi_deriv(a,b,m,r);
      delta = f/(fp-f*s);
      r -= delta;
      if (abs( delta ) < eps) {
	break;
      }
      else {
	j++;
      }
    }
    xs(k) = r;
  }

  return xs;
}


// computes (x,w) where x is a vector of quadrature points
// w is a vector of quadrature weights.  a and b are the
// weights of the jacobi polynomials
// Algorithm taken from Karniadakis & Sherwin, Appendix B
static Array<double,1>
gauss_jacobi_weights( double a, double b , int m )
{
  Array<double,1> xs = gauss_jacobi_points(a,b,m);
  Array<double,1> ws( xs.size() );
  
  double a1 = pow(2,a+b+1);

  // gamma seems to return ln(gamma)?  What platforms
  // does this happen on?
  double a2 = exp(gamma(a+m+1));
  double a3 = exp(gamma(b+m+1));
  double a4 = exp(gamma(a+b+m+1));
  double a5 = factorial(m);
  double a6 = a1*a2*a3/a4/a5;

  for (int k=0;k<m;k++) {
    double fp = Jacobi::eval_jacobi_deriv(a,b,m,xs(k));
    ws(k) = a6/(1-pow(xs(k),2)) / (pow(fp,2));
  }

  return ws;
}

static Array<double,2>
collapsed_points_triangle( int m )
{
  Array<double,1> ptx = gauss_jacobi_points(0,0,m);
  Array<double,1> pty = gauss_jacobi_points(1,0,m);
  Array<double,2> pts(ptx.rows()*pty.rows(),2);

  coord_transform xi = Expansions::inverse_coord_change( Shapes::TRIANGLE );

  int cur = 0;
  for (int i=0;i<ptx.size();i++) {
    Array<double,1> pt_in(2);
    for (int j=0;j<pty.size();j++) {
      pt_in(0) = ptx(i); pt_in(1) = pty(j);
      pts(cur++,Range::all()) = xi(pt_in);
    }
  }

  return pts;
}

static Array<double,2>
collapsed_points_tetrahedron( int m )
{
  Array<double,1> ptx = gauss_jacobi_points(0,0,m);
  Array<double,1> pty = gauss_jacobi_points(1,0,m);
  Array<double,1> ptz = gauss_jacobi_points(2,0,m);
  Array<double,2> pts(ptx.size()*pty.size()*ptz.size(),3);

  coord_transform xi = Expansions::inverse_coord_change( Shapes::TETRAHEDRON );

  int cur = 0;
  for (int i=0;i<ptx.size();i++) {
    Array<double,1> pt_in(3);
    for (int j=0;j<pty.size();j++) {
      for (int k=0;k<ptz.size();k++) {
	pt_in(0) = ptx(i); pt_in(1) = pty(j); pt_in(2) = ptz(k);
	pts(cur++,Range::all()) = xi( pt_in );
      }
    }
  }
  return pts;
  
}

static Array<double,1>
collapsed_weights_triangle( int m )
{
  Array<double,1> wx = gauss_jacobi_weights(0,0,m);
  Array<double,1> wy = gauss_jacobi_weights(1,0,m);
  Array<double,1> ws( wx.size() * wy.size() );

  int cur=0;
  for (int i=0;i<wx.rows();i++) {
    for (int j=0;j<wy.rows();j++) {
      ws(cur++) = 0.5 * wx(i) * wy(j);
    }
  }

  return ws;
}

static Array<double,1>
collapsed_weights_tetrahedron( int m )
{
  Array<double,1> wx = gauss_jacobi_weights(0,0,m);
  Array<double,1> wy = gauss_jacobi_weights(1,0,m);
  Array<double,1> wz = gauss_jacobi_weights(2,0,m);
  Array<double,1> ws( wx.size() * wy.size() * wz.size() );

  int cur=0;
  for (int i=0;i<wx.rows();i++) {
    for (int j=0;j<wy.rows();j++) {
      for (int k=0;k<wz.rows();k++) {
	ws(cur++) = 0.125 * wx(i) * wy(j) *wz(k);
      }
    }
  }
  return ws;
}

JacobiQuadratureRule::JacobiQuadratureRule( double a , double b , int m ) :
  QuadratureRule( Array<double,2>( gauss_jacobi_points(a,b,m).data() ,
				   shape(m,1) , duplicateData ) ,
		  gauss_jacobi_weights(a,b,m) ) {}

CollapsedQuadratureRuleTriangle::CollapsedQuadratureRuleTriangle( int m ):
  QuadratureRule( collapsed_points_triangle(m) ,
		  collapsed_weights_triangle(m) ) {}


CollapsedQuadratureRuleTetrahedron::CollapsedQuadratureRuleTetrahedron( int m ):
  QuadratureRule( collapsed_points_tetrahedron(m) ,
		  collapsed_weights_tetrahedron(m) ) {}

namespace FIAT {
RCP<QuadratureRule> make_collapsed_quadrature( int shp , int m )
{
  if (shp == Shapes::LINE) {
    RCP<QuadratureRule> rule = new GaussQuadratureRule( m );
    return rule;
  }
  else if (shp == Shapes::TRIANGLE) {
    RCP<QuadratureRule> rule = new CollapsedQuadratureRuleTriangle( m );
    return rule;
  }
  else if (shp == Shapes::TETRAHEDRON) {
    RCP<QuadratureRule> rule = new CollapsedQuadratureRuleTetrahedron( m );
    return rule;
  }
  else FIAT_ERROR("invalid shape " << shp << " in make_collapsed_quadrature()");
}
}
