// Written by Robert C. Kirby
// Copyright 2005 by The University of Chicago
// Distributed under the LGPL license
// This work is partially supported by the US Department of Energy
// under award number DE-FG02-04ER25650
#include "FIAT_shapes.hpp"
#include <iostream>


using namespace FIAT;
BZ_USING_NAMESPACE(blitz);


typedef int (*int2int)(int);

static const int pt_num_ents[] = {1};
static const int line_num_ents[] = {2,1};
static const int tri_num_ents[] = {3,3,1};
static const int tet_num_ents[] = {4,6,4,1};

static const int *num_ents[] = {pt_num_ents,
				line_num_ents,
				tri_num_ents,
				tet_num_ents};

// vertex coordinates for each reference element;
static double v0_line[] = {-1.0};
static double v1_line[] = {1.0};

static Array<double,1> v0_line_b( v0_line , shape(1) ,
				  duplicateData );
static Array<double,1> v1_line_b( v1_line , shape(1) ,
				  duplicateData);

static double v0_tri[] = {-1.0,-1.0};
static double v1_tri[] = {1.0,-1.0};
static double v2_tri[] = {-1.0,1.0};

static Array<double,1> v0_tri_b( v0_tri , shape(2) ,
				 duplicateData );
static Array<double,1> v1_tri_b( v1_tri , shape(2) ,
				 duplicateData );
static Array<double,1> v2_tri_b( v2_tri , shape(2) ,
				 duplicateData );


static double v0_tet[] = {-1.0,-1.0,-1.0};
static double v1_tet[] = {1.0,-1.0,-1.0};
static double v2_tet[] = {-1.0,1.0,-1.0};
static double v3_tet[] = {-1.0,-1.0,1.0};


static double *vertex_coords_line[] = { v0_line ,
					v1_line };

static double *vertex_coords_tri[] = { v0_tri ,
				       v1_tri ,
				       v2_tri };

static double *vertex_coords_tet[] = { v0_tet ,
				       v1_tet ,
				       v2_tet ,
				       v3_tet };

static double **vertex_coords[] = { NULL,
				    vertex_coords_line ,
				    vertex_coords_tri ,
				    vertex_coords_tet };

// connectivity
// map from edges of shapes to vertices comprising those edges
static int e0_line[] = {0,1};

static int e0_tri[] = {1,2};
static int e1_tri[] = {2,0};
static int e2_tri[] = {0,1};

static int e0_tet[] = {1,2};
static int e1_tet[] = {2,0};
static int e2_tet[] = {0,1};
static int e3_tet[] = {0,3};
static int e4_tet[] = {1,3};
static int e5_tet[] = {2,3};

static int *line_edges[] = { e0_line };
static int *triangle_edges[] = { e0_tri , e1_tri , e2_tri };
static int *tetrahedron_edges[] = { e0_tet , e1_tet , e2_tet ,
				    e3_tet , e4_tet , e5_tet };

static int **edges[] = { line_edges ,
			 triangle_edges , 
			 tetrahedron_edges };

// map from face id to vertices on the tetrahedron
static int f0_tri[] = {0,1,2};

static int f0_tet[] = {1,3,2};
static int f1_tet[] = {2,3,0};
static int f2_tet[] = {3,1,0};
static int f3_tet[] = {0,1,2};

static int *triangle_faces[] = {f0_tri};

static int *tetrahedron_faces[] = { f0_tet , f1_tet ,
				  f2_tet , f3_tet };

static int **faces[] = { triangle_faces , tetrahedron_faces };

static int t0_tet[] = {0,1,2,3};
static int *tetrahedron_tets[] = {t0_tet};

static int **vertex_relations_line[] = 
  { line_edges };

static int **vertex_relations_triangle[] =
  { NULL , triangle_edges , triangle_faces };

static int **vertex_relations_tetrahedron[] =
  { NULL , tetrahedron_edges , tetrahedron_faces , tetrahedron_tets };

static int ***vertex_relations[] =
  { NULL , vertex_relations_line ,
    vertex_relations_triangle ,
    vertex_relations_tetrahedron };

static int poly_dim_line( int n )
{
  if (n<0) throw exception();
  return n+1;
}

static int poly_dim_tri( int n )
{
  if (n<0) throw exception();
  return (n+1)*(n+2)/2;
}

static int poly_dim_tet( int n )
{
  if (n<0) throw exception();
  return (n+1)*(n+2)*(n+3)/6;
}

static int2int pdims[] = {NULL,poly_dim_line,poly_dim_tri,poly_dim_tet};

// normal vectors
static const double M_SQRT1_3 = 1.0/sqrt(3.0);
static const double nl0[] = {-1.0};
static const double nl1[] = {1.0};
static const double *normals_line[] = { nl0 , nl1 };

static const double ntri0[] = {M_SQRT1_2,M_SQRT1_2};
static const double ntri1[] = {-1.0,0.0};
static const double ntri2[] = {0.0,-1.0};
static const double *normals_tri[] = {ntri0,ntri1,ntri2};

static const double ntet0[] = {M_SQRT1_3,M_SQRT1_3,M_SQRT1_3};
static const double ntet1[] = {-1.,0.,0.};
static const double ntet2[] = {0.,-1.,0.};
static const double ntet3[] = {0.,0.,-1.};
static const double *normals_tet[] = {ntet0,ntet1,ntet2,ntet3};

static const double **normals[] = { NULL ,
				    normals_line ,
				    normals_tri ,
				    normals_tet };

// tangent vectors
static double ttri0[] = {-M_SQRT1_2,M_SQRT1_2};
static double ttri1[] = {0,-1};
static double ttri2[] = {1,0};
static double *tangents_tri[] = {ttri0,ttri1,ttri2};

static double ttet0[] = {-M_SQRT1_2,M_SQRT1_2,0};
static double ttet1[] = {0,-1,0};
static double ttet2[] = {1,0,0};
static double ttet3[] = {0,0,1};
static double ttet4[] = {-M_SQRT1_2,0,M_SQRT1_2};
static double ttet5[] = {0,-M_SQRT1_2,M_SQRT1_2};
static double *tangents_tet[] = {ttet0,ttet1,ttet2,ttet3,ttet4,ttet5};

static double **tangents[] = {NULL,NULL,tangents_tri,tangents_tet};


int Shapes::polynomial_dimension( int shp , int n )
{
  if (shp<1 || shp>3) throw exception();
  return pdims[shp](n);
};

int Shapes::topological_dimension( int shp ) {
  if ( shp < 0 || shp > 3 ) {
    throw exception();
  }
  return shp;
}

int Shapes::num_entities( int shp , int dim ) {
  if ( dim < 0 || dim > shp || shp < 0 || shp > 3 ) {
    throw exception();
  }
  return num_ents[shp][dim];
}

Array<double,1> Shapes::get_vertex( int shp , int v_id )
{
  const int d = topological_dimension( shp );
  const double *v = vertex_coords[shp][v_id];
  Array<double,1> x(d);
  for (int i=0;i<d;i++) x(i) = v[i];
  return x;
}

Array<int,1> Shapes::get_vertex_ids( int shp ,
				     int ent_dim ,
				     int ent_id )
{
  if (shp < 1 || shp > 3) throw exception();
  if (ent_id >= num_entities( shp , ent_dim )) throw exception();

  if (ent_dim == 0) {
    Array<int,1> vids(1);
    vids(0) = ent_id;
    return vids;
  }
  else {
    Array<int,1> vids(vertex_relations[shp][ent_dim][ent_id],
		      shape(ent_dim+1),duplicateData);
    return vids;
  }
}

static Array<double,2>
make_lattice_line( int n ,
		   int interior )
{
  int np = 0;
  for (int jj=interior;jj<n+1-interior;jj++) np++;
  Array<double,2> xs(np,1);

  if (np == 0) return xs;

  const double *v0 = vertex_coords[Shapes::LINE][0];
  const double *v1 = vertex_coords[Shapes::LINE][1];
  double h = abs( *v1 - *v0 );

  int cur=0;
  for (int jj=interior;jj<n+1-interior;jj++) {
    xs(cur,0) = *v0 + ( h*jj ) / n;
    cur++;
  }

  return xs;
}

static Array<double,2> make_lattice_triangle( int n ,
					      int interior )
{
  int np = 0;
  for (int ii=interior;ii<n+1-interior;ii++) {
    for (int jj=interior;jj<n+1-ii-interior;jj++) {
      np++;
    }
  }

  Array<double,2> xs(np,2);
  if (np==0) return xs;
  
  double **vs = vertex_coords[Shapes::TRIANGLE];
  double hs[2][2];
  for (int i=0;i<2;i++) {
    for (int j=0;j<2;j++) {
      hs[i][j] = vs[i+1][j]-vs[0][j];
    }
  }

  int cur_pt = 0;
  for (int ii=interior;ii<n+1-interior;ii++) {
    for (int jj=interior;jj<n+1-ii-interior;jj++) {
      for (int j=0;j<2;j++) {
	xs(cur_pt,j) = vs[0][j]+ii*hs[1][j]/n+jj*hs[0][j]/n;
      }
      cur_pt++;
    }
  }

  return xs;
}

static Array<double,2> make_lattice_tetrahedron( int n ,
						 int interior)
{
  int np = 0;
  for (int ii=interior;ii<n+1-interior;ii++) {
    for (int jj=interior;jj<n+1-ii-interior;jj++) {
      for (int kk=interior;kk<n-ii-jj+1-interior;kk++) {
	np++;
      }
    }
  }

  Array<double,2> xs(np,3);
  if (np == 0) return xs;

  double **vs = vertex_coords[Shapes::TETRAHEDRON];
  double hs[3][3];
  for (int i=0;i<3;i++) {
    for (int j=0;j<3;j++) {
      hs[i][j] = vs[i+1][j]-vs[0][j];
    }
  }

  int cur_pt = 0;
  for (int ii=interior;ii<n+1-interior;ii++) {
    for (int jj=interior;jj<n+1-ii-interior;jj++) {
      for (int kk=interior;kk<n-ii-jj+1-interior;kk++) {
	for (int j=0;j<3;j++) {
	  xs(cur_pt,j) = vs[0][j]
	    +ii*hs[2][j]/n
	    +jj*hs[1][j]/n
	    +kk*hs[0][j]/n;
	}
	cur_pt++;
      }
    }
  }
  
  return xs;
}

typedef Array<double,2> (*lat_func)(int,int);

static lat_func lattice_functions[] = { NULL,
					make_lattice_line ,
					make_lattice_triangle ,
					make_lattice_tetrahedron };

Array<double,2> Shapes::make_lattice( int shp ,
				      int n ,
				      int interior )
{
  return lattice_functions[shp](n,interior);
}

//  verts is the tuple of vertex ids
PtToEdge::PtToEdge( int shp , const Array<int,1> &verts ) :
  a0( vertex_coords[1][0][0] ) , b0( vertex_coords[1][1][0] ) ,
  h( 1./(vertex_coords[1][1][0] - vertex_coords[1][0][0]) ) , 
  shp(shp) ,
  v0( vertex_coords[shp][verts(0)] ,
      shape(Shapes::topological_dimension(shp)) , 
      duplicateData  ) ,
  v1( vertex_coords[shp][verts(1)] ,
      shape(Shapes::topological_dimension(shp)) , 
      duplicateData  ) ,
  diff(Shapes::topological_dimension(shp))
{
  diff = v1 - v0;
}

void PtToEdge::operator()(const Array<double,1> &xin,
			  Array<double,1> &xout) const
{
  int d = Shapes::topological_dimension( shp );
  for (int i=0;i<d;i++) {
    xout(i) = v0(i)+h*(xin(0)-a0)*diff(i);
  }
  return;
}

PtToFaceTet::PtToFaceTet( const Array<int,1> &verts ) :
  v0( vertex_coords[3][verts(0)] , shape(3) , duplicateData ) ,
  v1( vertex_coords[3][verts(1)] , shape(3) , duplicateData ) ,
  v2( vertex_coords[3][verts(2)] , shape(3) , duplicateData ) ,
  d0( 3 ) , d1( 3 )
{
  d0 = v1 - v0;
  d1 = v2 - v0;
}

void PtToFaceTet::operator()(const Array<double,1> &xin ,
			     Array<double,1> &xout ) const
{
  xout = v0 + 0.5 * ( xin(0) + 1.0 ) * d0 + 0.5 * (xin(1)+1.0) * d1;
  return;
}

static IdentityPtMap ident;

static PtMap *pt_to_edge_edge[] = {&ident};

static PtToEdge pte_tri[] =
  { PtToEdge(2,Shapes::get_vertex_ids(2,1,0)),
    PtToEdge(2,Shapes::get_vertex_ids(2,1,1)),
    PtToEdge(2,Shapes::get_vertex_ids(2,1,2)) };

static PtMap *pt_to_edge_tri[] =
  { &pte_tri[0] , &pte_tri[1] , &pte_tri[2] };

static PtToEdge pte_tet[] =
  { PtToEdge(3,Shapes::get_vertex_ids(3,1,0)) ,
    PtToEdge(3,Shapes::get_vertex_ids(3,1,1)) ,
    PtToEdge(3,Shapes::get_vertex_ids(3,1,2)) ,
    PtToEdge(3,Shapes::get_vertex_ids(3,1,3)) ,
    PtToEdge(3,Shapes::get_vertex_ids(3,1,4)) ,
    PtToEdge(3,Shapes::get_vertex_ids(3,1,5)) };

static PtMap *pt_to_edge_tet[] =
  { &pte_tet[0] , &pte_tet[1] , &pte_tet[2] ,
    &pte_tet[3] , &pte_tet[4] , &pte_tet[5] };

static PtMap *pt_to_face_tri[] = { &ident };

static PtToFaceTet ptf_tet[] =
  { PtToFaceTet( Shapes::get_vertex_ids(3,2,0) ) ,
    PtToFaceTet( Shapes::get_vertex_ids(3,2,1) ) ,
    PtToFaceTet( Shapes::get_vertex_ids(3,2,2) ) ,
    PtToFaceTet( Shapes::get_vertex_ids(3,2,3) ) };

static PtMap *pt_to_face_tet[] = 
  { &ptf_tet[0] , &ptf_tet[1] ,
    &ptf_tet[2] , &ptf_tet[3] ,
    &ptf_tet[4] , &ptf_tet[5] };

static PtMap *pt_to_tet_tet[] = { &ident };

static PtMap **pt_maps_line[] = { NULL, pt_to_edge_edge };
static PtMap **pt_maps_tri[] = { NULL , pt_to_edge_tri ,
			       pt_to_face_tri };
static PtMap **pt_maps_tet[] = { NULL , pt_to_edge_tet ,
				 pt_to_face_tet ,
				 pt_to_tet_tet };

static PtMap ***pt_maps[] = { NULL , pt_maps_line , pt_maps_tri, pt_maps_tet};

PtMap &Shapes::get_pt_map( int shp ,
			   int ent_dim ,
			   int ent_id )
{
  return *pt_maps[shp][ent_dim][ent_id];
}

Array<double,2> Shapes::make_points( int shp ,
				     int ent_dim ,
				     int ent_id ,
				     int n )
{
  int d = topological_dimension( shp );

  if (ent_dim == 0) { 
    const double *v = vertex_coords[shp][ent_id];
    Array<double,2> pts(1,d);
    for (int i=0;i<d;i++) pts(0,i) = v[i];
    return pts;
  }
  if (ent_dim == d) {
    if (ent_id != 0) throw exception();
    return make_lattice( shp , n , 1 );
  }
  else {
    PtMap &f = *pt_maps[shp][ent_dim][ent_id];
    Array<double,2> pre_xs = make_lattice(ent_dim , n , 1);
    Array<double,2> xs(pre_xs.rows(),d);

    for (int i=0;i<pre_xs.rows();i++) {
      Array<double,1> pre_xcur = pre_xs(i,Range::all());
      Array<double,1> xcur = xs(i,Range::all());
      f(pre_xcur,xcur);
    }
    return xs;
  }

}

// vector<double> Shapes::get_normal( int shape ,
// 				   int ent )
// {
//   const int d = topological_dimension( shape );
//   if (shape < 1 || shape > 3) throw exception();
//   if (ent >= num_entities( shape , d-1 )) throw exception();
//   const double *n = normals[shape][ent];
//   return vector<double>(n,n+d);
// }

// vector<double> Shapes::get_tangent( int shape ,
// 				    int ent )
// {
//   if (shape < 2 || shape > 3) throw exception();
//   if (ent >= num_entities(shape,1)) throw exception();
//   const double *t = tangents[shape][ent];
//   const int d = topological_dimension( shape );
//   return vector<double>(t,t+d);
// }
